"""
ElectroSense Controller
(c) ElectroSense Project

Markus Fuchs - fuchs@sero-systems.de - 2019-2020
"""
import asyncio

from dataclasses import dataclass, field, is_dataclass
from enum import Enum, unique
from typing import Any, Tuple, Optional, Dict, List

from .sensorinfo import SensorInfo
from .taskinfo import TaskInfo


@unique
class RTCState(Enum):
    """State of an RTC session between a browser and a sensor"""
    NONE = 0
    USER_REQUESTED_OFFER = 1
    SENSOR_SENT_OFFER = 2
    USER_SENT_ANSWER = 3
    # session is established and active, data flows between browser and sensor
    ICE_CANDIDATES_EXCHANGE = 4


@dataclass(eq=True, frozen=True)
class SignalingClient:
    """A client connected to the signaling server"""
    websocket: Any


@dataclass(eq=True, frozen=True)
class SignalingBrowser(SignalingClient):
    """A client of the signaling server which identified as user in a web browser"""
    user_id: str


@dataclass(eq=True, frozen=True)
class SignalingSensor(SignalingClient):
    """A sensor connected to the signaling server"""
    sensor_id: str
    reported_state: str = ""


@dataclass(eq=True, frozen=True)
class RTCSession:
    """A session between a user in a browser and a sensor"""
    browser: SignalingBrowser
    sensor: SignalingSensor
    state: RTCState = RTCState.NONE
    start_time: int = 0
    end_time: int = 0


@dataclass(eq=True, frozen=True)
class SensorState:
    """A sensor's current state"""
    serial: Any = None
    signaling: Optional[SignalingSensor] = None
    mqtt: Optional[SensorInfo] = None


@dataclass(eq=True, frozen=True)
class MqttState:
    connected: bool = False


@dataclass(eq=True, frozen=True)
class MqttMessage:
    msg_type: str = ""
    msg: Dict = field(default_factory=dict)
    topic: str = ""


@dataclass(eq=True, frozen=True)
class CampaignAllocation:
    sensors: Tuple[SensorInfo] = field(default_factory=tuple)
    task: Optional[TaskInfo] = None
    # true to start the task, false to stop
    start: bool = True


@dataclass()
class AppState:
    """The complete application state"""
    rtc_sessions: Tuple[RTCSession, ...] = field(default_factory=tuple)
    # browsers connected to the signaling server over web socket
    ws_browsers: Dict[str, SignalingBrowser] = field(default_factory=dict)
    # sensors connected to the signaling or MQTT server
    sensors: Dict[str, SensorState] = field(default_factory=dict)
    # unknown clients connected over web socket
    unknown_ws_clients: Tuple[SignalingClient, ...] = field(default_factory=tuple)

    # state of the MQTT connection
    mqtt_state: MqttState = field(default_factory=MqttState)
    # tasks/campaigns
    mqtt_tasks: Tuple[TaskInfo, ...] = field(default_factory=tuple)

    # scheduled future tasks
    future_tasks: Tuple[Tuple[str, asyncio.Task], ...] = field(default_factory=tuple)


class MqttObserver:
    def on_mqtt_connect(self):
        pass

    def on_mqtt_disconnect(self):
        pass

    def on_mqtt_sensor_log(self, msg: Dict):
        pass

    def on_mqtt_sensor_info(self, msg: Dict):
        pass

    def on_mqtt_sensor_add(self, sensor: SensorInfo):
        pass

    def on_mqtt_sensor_remove(self, sensor: str):
        pass

    def on_mqtt_sensor_status(self, msg: Dict):
        pass

    def on_mqtt_task_recover(self, tasks: List[TaskInfo]):
        pass

    def on_mqtt_task_added(self, task: TaskInfo) -> Optional[int]:
        pass

    def on_mqtt_task_removed(self, task_id: int):
        pass

    def on_mqtt_sensor_status_poll(self):
        pass

    def get_state(self) -> AppState:
        pass


class ElectroSenseController(MqttObserver):
    def get_state(self) -> AppState:
        pass

    def get_active_rtc_sessions(self):
        pass

    def get_connected_signaling_users(self) -> List[str]:
        pass

    def get_connected_signaling_sensors(self) -> List[str]:
        pass

    def get_rtc_time_by_user(self, start: int = None, end: int = None) -> Optional[List[Dict[str, Any]]]:
        pass

    def get_rtc_time_by_sensor(self, start: int = None, end: int = None) -> Optional[List[Dict[str, Any]]]:
        pass

    def get_rtc_hourly_usage(self, start: int = None, end: int = None) -> Optional[List[Dict[str, Any]]]:
        pass


def clone_frozen(obj, **kwargs):
    """Clone an instance of a frozen dataclass while only changing certain
    attributes.

    Example:
      @dataclass(frozen=True, eq=True)
      class Foo:
          param1: int
          param2: str

      a = Foo(1, "test")
      b = clone_frozen(state, param2="changed")
      b == Foo(1, "changed")
      > True

    :param obj: instance to clone
    :param kwargs: parameters to be changed in the new instance
    """
    if not is_dataclass(obj):
        raise RuntimeError("Input must be a dataclass")

    for k, v in obj.__dict__.items():
        if k not in kwargs.keys():
            kwargs[k] = v

    return obj.__class__(**kwargs)


class IllegalStateError(Exception):
    """raised when application state is wrong"""
    pass