# -*- coding: utf-8 -*-
"""
Created on 2019-03-07

@author: hector
"""
import json
import logging
import re

from datetime import datetime, timezone
from dateutil import parser as dateparser
from enum import Enum
from typing import List, Tuple, Optional


logger = logging.getLogger(__name__)


# Responses
TASK_PARSE_OK = 0
TASK_PARSE_FAIL = 1

# Statuses
TASK_STATUS_UNINITIALIZED = 0
TASK_STATUS_PENDING = 1
TASK_STATUS_RUNNING = 2
TASK_STATUS_DONE = 3
TASK_STATUS_SKIPPED = 4


class TaskTimeType(Enum):
    ISO8601 = 0
    NOW = 1
    # until stopped by controller
    UNTIL_STOP = 2
    # until executed task on sensor has finished
    UNTIL_EXECUTION = 3


def transform_date(txt) -> Tuple[TaskTimeType, Optional[datetime]]:
    """Receives an text with a date in ISO8601 or a keyword.
       Returns a pair of date type and timestamp"""
    if txt == "NOW":
        return TaskTimeType.NOW, None
    if txt == "UE":
        return TaskTimeType.UNTIL_EXECUTION, None
    if txt == "US":
        return TaskTimeType.UNTIL_STOP, None
    return TaskTimeType.ISO8601, dateparser.parse(txt)


def is_naive(ts) -> bool:
    """Returns true if the datetime object is not timezone aware."""
    return ts.tzinfo is None or ts.tzinfo.utcoffset(ts) is None


class TaskParseException(Exception):
    def __init__(self, errors: List[str]):
        self.errors = errors


class TaskInfo:
    def __init__(self):
        self.issuer = None
        self.sensors = set()  # As defined in the campaign unless UE
        self.priority = None
        self.start = None
        self.start_stamp = None
        self.stop = None
        self.stop_stamp = None
        self.parameters = None
        self.id = None
        self.start_timer = None
        self.stop_timer = None
        self.status = TASK_STATUS_UNINITIALIZED

    def to_json(self):
        """Returns a JSON representation of the task attributes"""
        return json.dumps(self.to_dict())

    def to_dict(self):
        """Returns a dictionary representation of the task attributes"""
        output = {'sensors': list(self.sensors),
                  'issuer': self.issuer,
                  'parameters': self.parameters,
                  'start': self.start_stamp.isoformat() if self.start == TaskTimeType.ISO8601 else "NOW",
                  'status': task_status_text(self.status),
                  'priority': self.priority}

        if self.stop == TaskTimeType.ISO8601:
            output['stop'] = self.stop_stamp.isoformat()
        if self.stop == TaskTimeType.UNTIL_EXECUTION:
            output['stop'] = "UE"
        if self.stop == TaskTimeType.UNTIL_STOP:
            output['stop'] = "US"

        return output

    def __repr__(self):
        t = self.to_dict()
        t["id"] = self.id
        return t.__repr__()

    def set_id(self, tid):
        self.id = tid
        self.parameters["CampaignId"] = str(tid)

    def is_valid(self):
        """True if task has valid parameters, False otherwise.
        :returns (valid, msgs): tuple of validity and error messages.
                                Error messages are only set if the task is invalid."""
        # Sanity checks
        valid = True
        msgs = []

        if self.start == TaskTimeType.UNTIL_EXECUTION or self.start == TaskTimeType.UNTIL_STOP:
            msgs.append("Start time is invalid: " + self.start)
            valid = False

        if self.start == TaskTimeType.ISO8601:
            if is_naive(self.start_stamp):
                msgs.append("Start time is not timezone-aware: " + str(self.start_stamp))
                valid = False

        if self.start_stamp == TaskTimeType.NOW:
            msgs.append("Stop time is invalid: " + self.stop)
            valid = False

        if self.stop == TaskTimeType.ISO8601:
            if is_naive(self.stop_stamp):
                msgs.append("Stop time is not timezone-aware: " + str(self.start_stamp))
                valid = False

        if self.stop == TaskTimeType.ISO8601 and self.start == TaskTimeType.ISO8601 and \
                self.start_stamp > self.stop_stamp:
            msgs.append("Task starts after stop.")
            valid = False

        if self.stop == TaskTimeType.ISO8601 and self.stop_stamp < datetime.now(timezone.utc):
            msgs.append("Task stops in the past")
            valid = False

        for _, v in self.parameters.items():
            if type(v) is not str and type(v) is not int:
                msgs.append("Parameters badly specified")
                valid = False
                break

        if "command" not in self.parameters:
            msgs.append("Task does not specify a command")
            valid = False
        else:
            if '/' in self.issuer or ' ' in self.issuer:
                msgs.append("Command badly specified")
                valid = False

        for s in self.sensors:
            if type(s) is not str and type(s) is not int:
                msgs.append("Some sensor element is incorrect.")
                valid = False
                break

        if type(self.issuer) is not str:
            msgs.append("Issuer badly specified.")
            valid = False
        else:
            if not re.match('^[a-zA-Z0-9_]+$', self.issuer):
                msgs.append("Issuer badly specified (only alphanumeric and _ allowed)")
                valid = False

        if self.status == TASK_STATUS_UNINITIALIZED:
            msgs.append("Task not initialized")
            valid = False

        return valid, msgs

    def has_collisions(self, other_tasks):
        """Check if TaskInfo does not have collisions with other_tasks with same priority and time.
        :returns (task_id, sensors): task_id of colliding task and sensors concerned as list.
                                     task_id is None if there are no collisions."""
        for check_task in other_tasks:
            if self.priority != check_task.priority:
                # has different priority
                continue  # Different priorities allowed in any case
            if (len(self.sensors) == 0 and len(self.sensors) != len(check_task.sensors)) \
               or len(self.sensors & check_task.sensors) < 1:
                # no overlap in affected sensors
                continue
            if is_task_time_clash(self, check_task):
                return check_task.id, self.sensors & check_task.sensors

        # Tasks starting NOW will not be accepted if they cannot run (think webrtc sessions)
        # If there exists a task running now that affects any of the new task's sensors
        # with higher priority it will be rejected
        # TODO [Markus] don't really get it --> probably simplify
        if self.start == TaskTimeType.NOW:
            candidate_tasks = [x for x in other_tasks
                               if x.status == TASK_STATUS_RUNNING
                               and x.priority > self.priority]
            for candidate in candidate_tasks:
                # If there is a clash in sensors (beware default tasks)
                if self.sensors & candidate.sensors or not candidate.sensors:
                    logger.debug("New task set to start NOW cannot proceed "
                                 "because there is a higher-priority task running affecting "
                                 "one or more sensors.")
                    return candidate.id, [0] if not candidate.sensors else self.sensors & candidate.sensors

        return None, []


def from_json(message):
    """Parses a text string which JSON-encodes a task definition."""
    try:
        task = json.loads(message)
    except json.JSONDecodeError:
        raise TaskParseException(["Invalid JSON for the task: " + str(message)])
    return from_dict(task)


def from_dict(task):
    """Parses a dictionary which encodes a task definition."""
    try:
        task_sensors = task['sensors']  # List of sensors
        task_issuer = task['issuer']  # Executable to run
        task_parameters = task['parameters']  # Dictionary of parameters and values
        task_start, task_start_stamp = transform_date(task['start'])  # NOW or ISO8601
        task_stop, task_stop_stamp = transform_date(task['stop'])  # UE, US or ISO8601
        task_priority = int(task['priority'])  # Numeric priority
    except KeyError as e:
        raise TaskParseException(["Task JSON not well defined, lacking: {}.".format(str(e))])
    except (ValueError, OverflowError) as e:
        raise TaskParseException(["Task uses badly formatted date or priority: {}".format(str(e))])

    if not isinstance(task_parameters, dict):
        raise TaskParseException(["Parameters are not a dict"])

    if not isinstance(task_sensors, list):
        raise TaskParseException(["Sensors are not a list"])

    ti = TaskInfo()
    ti.issuer = task_issuer
    ti.sensors = set([str(s) for s in task_sensors])
    ti.priority = task_priority
    ti.start = task_start
    ti.start_stamp = task_start_stamp
    ti.stop = task_stop
    ti.stop_stamp = task_stop_stamp
    ti.parameters = {}

    # Define all parameters as (opaque) strings we will not process
    #  Fixing everything to string helps in JSON (REST/DB)<->Avro (over MQTT) interactions
    for parameter in task_parameters:
        ti.parameters[parameter] = str(task_parameters[parameter])
    ti.status = TASK_STATUS_PENDING

    valid, errs = ti.is_valid()
    if not valid:
        raise TaskParseException(errs)

    return ti


def task_status_text(code):
    if code == TASK_STATUS_UNINITIALIZED:
        return "Uninitialized"
    if code == TASK_STATUS_PENDING:
        return "Pending"
    if code == TASK_STATUS_RUNNING:
        return "Running"
    if code == TASK_STATUS_DONE:
        return "Done"
    if code == TASK_STATUS_SKIPPED:
        return "Skipped"
    return "Unknown status code"


def is_task_time_clash(prev, new):
    """Check if two tasks share some time range"""
    # US and UE don't have an end time
    if (new.stop == TaskTimeType.UNTIL_EXECUTION or new.stop == TaskTimeType.UNTIL_STOP) and \
       (new.start == TaskTimeType.NOW or new.start_stamp > prev.start_stamp):
        return True

    if prev.start_stamp is None or new.start_stamp is None or prev.stop_stamp is None or new.stop_stamp is None:
        return False

    # Here all instants are timestamps
    return prev.start_stamp < new.start_stamp < prev.stop_stamp or prev.start_stamp < new.stop_stamp < prev.stop_stamp


def create_default_task():
    """Returns the default task (PSD 24-6000 MHz)"""
    task = TaskInfo()
    task.issuer = "default"
    task.priority = 1
    task.start, task.start_stamp = transform_date("NOW")
    task.stop, task.stop_stamp = transform_date("US")
    task.id = 1
    task.status = TASK_STATUS_PENDING

    parameters = {"command": "es_sensor",
                  "clkOffset": "0",
                  "devIndex": "0",
                  "log2FFTsize": "8",
                  "sslCollector": "collector.electrosense.org:5001#"
                                  "certs/CA-Cert.pem#certs/Sensor-SSL-Cert.pem#certs/Sensor-SSL-SK.pem",
                  "freqOverlap": "0.167",
                  "hoppingStrategy": "sequential",
                  "avgFactor": "5",
                  "minTimeRes": "0",
                  "window": "hanning",
                  "clkCorrPeriod": "3600",
                  "minfreq": "24000000",
                  "fftBatchLen": "10",
                  "maxfreq": "1766000000",      # 6000000000
                  "soverlap": "128",
                  "tcpCollector": "0",
                  "sampRate": "2400000",
                  "monitorTime": "0",
                  "gain": "32.8"}

    task.parameters = parameters

    return task
