# -*- coding: utf-8 -*-

"""
Controller v2.0 MQTT flavour
(c) Electrosense Project

[TODO] License text

Héctor Cordobés - hector.cordobes@imdea.org - 2017-2019

MQTT Module

"""

import logging
import sys
import os

import paho.mqtt.client as mqtt


logger = logging.getLogger(__name__)


class MqttEngine(object):

    def __init__(self, config, parser, topics):
        self.connection = False
        self.parser = parser
        self.topics = topics
        self.mc = None
        self.config = config
        self.actions = {}
        self.mqtt_on = lambda: None
        self.mqtt_off = lambda: None
        self.dev_mode = False
        self._setup()

    def register_actions(self, actions, mqtt_events=None):
        """register actions for certain messages and mqtt connect/disconnect events"""
        self.actions = actions

        if mqtt_events is None:
            return
        if "on" in mqtt_events.keys():
            self.mqtt_on = mqtt_events["on"]
        if "off" in mqtt_events.keys():
            self.mqtt_off = mqtt_events["off"]

    def register_last_will(self, topic, msg_type, msg):
        """Set message to be sent when uncontrolled disconnection occurs"""
        self.mc.will_set(topic, payload=self.parser.encode_message(msg_type, msg), qos=1)

    def _setup(self):
        """ Sets up the MQTT connection with the config data from init and starts."""
        # Configure MQTT connection (network, credentials and callbacks)
        self.dev_mode = self.config.get("dev_mode", False)

        self.mc = mqtt.Client(client_id=self.config["userid"],
                              clean_session=True,
                              userdata=None)

        self.mc.username_pw_set(username=self.config["user"],
                                password=self.config["password"])

        logger.debug("TLS files:\nCA " + self.config["ca"]
                     + "\nCert " + self.config["certificate"]
                     + "\nKey " + self.config["key"])

        self.mc.tls_set(ca_certs=self.config["ca"],
                        certfile=self.config["certificate"],
                        keyfile=self.config["key"],
                        tls_version=2)

        self.mc.on_connect = self.on_connect
        self.mc.on_message = self.on_mqtt_message
        self.mc.on_disconnect = self.on_disconnect

        self.mc.max_inflight_messages_set(int(self.config["inflight"]))
        self.mc.enable_logger(logger)

        try:
            self.mc.connect(self.config["server"], int(self.config["port"]), 60)
        except Exception as err:
            print("Could not connect to MQTT broker. Check MQTT configuration.")
            print(err)
            # Nuke controller
            sys.stdout.flush()
            os._exit(1)

    def on_connect(self, client, *_):  # userdata,flags,rc not used
        """ Callback to handle the connection to the MQTT broker."""
        logger.info("Connected to MQTT broker.")
        for topic in self.topics:
            logger.info("Subscribing to topic '%s'", topic)
            (r, _) = client.subscribe(topic, 1)
            if r == mqtt.MQTT_ERR_SUCCESS:
                logger.info("Subscription to topic '%s' succeeded", topic)
        self.connection = True
        self.mqtt_on()

    def on_mqtt_message(self, client, userdata, message):
        """ Callback to process a newly received MQTT message."""
        try:
            r = self.parser.decode_message(message.payload)
        except:
            logger.warning("Incorrect AVRO message received")
            return
        if "Type" not in r:
            logger.warning("Incorrect message received, no Type field")
            return
        mt = r["Type"]
        logger.debug("Received message type %s", mt)
        if mt not in self.actions:
            logger.warning("Incorrect message type received: %s", mt)
        # HACK clean, all serial numbers have to be strings (extensibility)
        if "SerialNumber" in r["Message"].keys():
            r["Message"]["SerialNumber"] = str(r["Message"]["SerialNumber"])

        # print("Received MQTT Message", str(r))
        self.actions[mt](r["Message"])

    def on_disconnect(self, client, userdata, rc):
        """ Callback when an MQTT disconnection from the broker is detected."""
        self.connection = False
        self.mqtt_off()
        if rc != 0:
            logger.warning("Unexpected disconnection, rc=%d.", rc)

    # Generic AVRO/MQTT message sending function
    def send_message(self, msg_type, data, topic):
        """From type and content dictionary, send completely
        formed AVRO messages over MQTT and log. """
        if self.dev_mode:
            logger.error("MQTT publish DISABLED in DEV MODE.")
            return
        msg = self.parser.encode_message(msg_type, data)
        (result, mid) = self.mc.publish(topic, msg, 0)
        logger.debug("Sending message with topic: %s, result: %d, id: %d", topic, result, mid)

    def start(self):
        """Start the MQTT main loop in a new thread"""
        self.mc.loop_start()

    def shutdown(self):
        """Stop the MQTT main loop thread"""
        self.mc.loop_stop()
        self.mc.disconnect()

    def is_alive(self):
        return self.mc.socket() is not None
